/*
** File: bitstrip.c - strip header from a .bit file and save only the SelectMap
**                    data.
**
** (C) Copyright Alpha Data 2012, 2014, 2016
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
#   if !defined(_tcserror)
#     /* For some reason, 32-bit MINGW doesn't have _tcserror */
#     define _tcserror strerror
#   endif
# endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tmain main
# define _tprintf printf
# define _tcserror strerror
# define _tfopen fopen
# define FALSE (0)
# define TRUE (1)

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <admxrc3.h>

#include <sdk_common.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define EXIT_OK                      (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE        (1) /* Illegal basic command-line syntax */
#define EXIT_INSUFFICIENT_ARGS       (2) /* Not enough positional arguments */
#define EXIT_LOAD_BIT_ERROR          (3) /* Failed to read input .bit file */
#define EXIT_SAVE_BIN_ERROR          (4) /* Failed to write output .bin file */

#ifdef _WIN32
# define PROGRAM_NAME _T("bitstrip.exe")
#else
# define PROGRAM_NAME _T("bitstrip")
#endif

#if !defined(_MSC_VER)
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = _tfopen(pFilename, pMode);
  if (NULL != pFile) {
    *ppFile = pFile;
    return 0;
  } else {
    return errno;
  }
}

static int
_tcserror_s(
  TCHAR* pBuffer,
  size_t n,
  int err)
{
  const TCHAR* p = _tcserror(err);
  size_t i;

  if (n == 0) {
    return 0;
  }
  if (NULL == pBuffer) {
    return EINVAL;
  }
  for (i = 0; i < n; i++) {
    pBuffer[i] = p[i];
    if (p[i] == _T('\0')) {
      break;
    }
  }
  pBuffer[n - 1] = _T('\0');
  return 0;
}
# endif

static int
doBitstrip(
  const TCHAR* pInputFilename,
  const TCHAR* pOutputFilename) /* May be NULL */
{
  ADMXRC3_BITSTREAM* pBitstream = NULL;
  ADMXRC3_STATUS status;
  FILE* pOutFile = NULL;
  int ret = EXIT_OK;
  
  if (NULL == pInputFilename) {
    _tprintf(_T("*** admxrc3Bitstrip requires input filename as first argument\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  status = ADMXRC3_LoadBitstream(pInputFilename, &pBitstream);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to load '%s': %s\n"), pInputFilename, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_LOAD_BIT_ERROR;
    goto done;
  }
  _tprintf(_T("'%s' -> device %s, SelectMap data length %llu(0x%llx) bytes.\n"),
    pInputFilename, pBitstream->Identifier, (unsigned long long)pBitstream->Length, (unsigned long long)pBitstream->Length);

  if (NULL != pOutputFilename) {
    size_t nWritten;
    int err;
    TCHAR buf[128];

    err = _tfopen_s(&pOutFile, pOutputFilename, _T("wb"));
    if (0 != err) {
      buf[0] = _T('\0');
      _tcserror_s(buf, ARRAY_LENGTH(buf), err);
      _tprintf(_T("*** Failed to open '%s' for writing: %s\n"), pOutputFilename, buf);
      ret = EXIT_SAVE_BIN_ERROR;
      goto done;
    }
    nWritten = fwrite(&pBitstream->Data[0], sizeof(uint8_t), pBitstream->Length, pOutFile);
    if (nWritten != pBitstream->Length) {
      _tprintf(_T("*** Only managed to write %llu(0x%llx) of %llu(0x%llx) bytes for file '%s'\n"),
        (unsigned long long)nWritten, (unsigned long long)nWritten,
        (unsigned long long)pBitstream->Length, (unsigned long long)pBitstream->Length,
        pOutputFilename);
      ret = EXIT_SAVE_BIN_ERROR;
      goto done;
    }
    fclose(pOutFile);
    pOutFile = NULL;
    _tprintf(_T("Wrote %llu(0x%llx) bytes to '%s'.\n"),
      (unsigned long long)nWritten, (unsigned long long)nWritten, pOutputFilename);
  }
  
done:
  if (NULL != pOutFile) {
    fclose(pOutFile);
  }
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstream(pBitstream);
  }
  
  return ret;
}

#if (defined(__VXWORKS__) || defined(__vxworks))
  
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/

int
admxrc3BitstripHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int admxrc3BitstripHelp()\n"));
  _tprintf(_T("  int admxrc3Bitstrip(pInputFilename[, pOutputFilename])\n"));
  _tprintf(_T("'pOutputFilename' may be NULL or omitted in order to display .bit file information only.\n"));
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
admxrc3Bitstrip(
  const TCHAR* pInputFilename,
  const TCHAR* pOutputFilename) /* May be NULL */
{
  return doBitstrip(pInputFilename, pOutputFilename);
}

#endif

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/
 
#if !(defined(__VXWORKS__) || defined(__vxworks))
int
_tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = EXIT_OK;
  Option options[1];
  OptionType paramType[1];
  TCHAR* paramInfo[1];
  TCHAR* paramMeaning[1];
  Arguments arguments = g_initArguments;
  TCHAR* pInputName;
  TCHAR* pOutputName;

  options[0].key = _T("o");
  options[0].type = OptionPath;
  options[0].def.pPathVal = _T("");
  options[0].meaning = _T("Output filename");
  paramType[0] = OptionPath;
  paramInfo[0] = _T("<input.bit>");
  paramMeaning[0] = _T(".bit file to be stripped");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = ARRAY_LENGTH(paramType);
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;

  if (sampleParseCommandLine(PROGRAM_NAME, argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }
  if (arguments.nParam < 1) {
    _tprintf(_T("*** input .bit filename must be specified as first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  pInputName = arguments.param[0].pPathVal;
  pOutputName = options[0].specified ? options[0].value.pPathVal : NULL;

  ret = doBitstrip(pInputName, pOutputName);

done:
  sampleCleanupCommandLine(&arguments);
  return ret;
}
#endif
