/*
** fpga.c - Infrastructure used by the Win32 sample applications for parsing
**          command-line arguments and generating filenames for .BIT files.
**
** (C) Copyright 2011-2013 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _vstprintf_s(buf, max, format, ...) _vsntprintf(buf, max, format, __VA_ARGS__)
#   define _tcscat_s(pDstStr, max, pStr) strcat(pDstStr, pStr)
#   define _tcscpy_s(pDstStr, max, pSrcStr) strcpy(pDstStr, pSrcStr)
#   define _tcslwr_s(pStr, max) strlwr_s(pStr, max)
# endif

#else

/* Linux or VxWorks */
# define _T(c) c
# if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
  /* VxWorks 5.x doesn't have snprintf() or vsnprintf() */
#   define _vstprintf_s(buf, max, format, ...) vsprintf(buf, format, __VA_ARGS__)
# else
#   define _vstprintf_s(buf, max, format, ...) vsnprintf(buf, max, format, __VA_ARGS__)
# endif
# define _tcscat_s(pDstStr, max, pStr) strcat(pDstStr, pStr)
# define _tcscpy_s(pDstStr, max, pSrcStr) strcpy(pDstStr, pSrcStr)
# define _tcslen(pStr) strlen(pStr)
# define _tcslwr_s(pStr, max) strlwr_s(pStr, max)
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "sdk_common.h"

#define NUL _T('\0')

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/*
** The environment variable that the sample applications
** attempt to pick up and use to find the base of the ADM-XRC
** SDK and hence the FPGA bitstream files.
*/
#define ENVIRONMENT_VARIABLE (_T("ADMXRC3_SDK"))

#if _WIN32
/* Windows */
static const TCHAR nativeSlashStr[] = _T("\\");
#else
/* Linux or VxWorks */
static const TCHAR nativeSlashStr[] = _T("/");
#endif

#define SET_STRING(STRING, VALUE) { \
  if ((STRING) != NULL) { \
		free(STRING); \
    (STRING) = NULL; \
  } \
  if ((VALUE) != NULL) { \
    size_t n = _tcslen(VALUE) + 1; \
    (STRING) = (TCHAR*)malloc(sizeof(TCHAR) * n); \
    memcpy((STRING), VALUE, sizeof(TCHAR) * n); \
  } \
}

#if !defined(_MSC_VER)
static void
strlwr_s(
  TCHAR* pStr,
  size_t max)
{
  TCHAR c;

  while (max) {
    c = *pStr;
    if (c == NUL) {
      break;
    }
    *pStr++ = (TCHAR)tolower(c);
    max--;
  }
}
#endif

#if !defined(_MSC_VER)
static void
_tgetenv_s(
  size_t* pRequiredLength,
  TCHAR* pBuffer,
  size_t max,
  const TCHAR* pVariableName)
{
  TCHAR* pTmp;
  size_t length;

  pTmp = getenv(pVariableName);
  if (NULL == pTmp) {
    length = 0;
  } else {
    length = strlen(pTmp);
  }
  if (NULL != pRequiredLength) {
    /* Not found */
    *pRequiredLength = length + 1;
  }
  if (NULL != pBuffer) {
    if (max) {
      max--; /* subtract 1 for NUL terminator */
      if (length > max) {
        length = max;
      }
      if (length) {
        memcpy(pBuffer, pTmp, length * sizeof(TCHAR));
      }
      pBuffer[length] = '\0';
    }
  }
}
#endif

TCHAR*
sampleGetSdkBitDir(
  void)
{
  const TCHAR bitDirName[] = _T("bit");
  size_t nEnvChar, nAlloc, dummy;
  TCHAR* pBitPath = NULL;
  
  _tgetenv_s(&nEnvChar, NULL, 0, ENVIRONMENT_VARIABLE);
  if (0 == nEnvChar) {
    return NULL;
  }
  
  nAlloc = nEnvChar + 1 + _tcslen(bitDirName) + 1;
  pBitPath = (TCHAR*)malloc(nAlloc * sizeof(TCHAR));
  if (NULL == pBitPath) {
    return NULL;
  }
  
  _tgetenv_s(&dummy, pBitPath, nAlloc, ENVIRONMENT_VARIABLE);
  _tcscat_s(pBitPath, nAlloc, nativeSlashStr);
  _tcscat_s(pBitPath, nAlloc, bitDirName);
  
  return pBitPath;
}

TCHAR*
sampleConstructBitFilename(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  const TCHAR* pDir,
  const TCHAR* pDesign,
  const TCHAR* pSdkBitDir,
  bool_t* pbShouldConfigure)
{
  const TCHAR ext[] = _T(".bit");
  TCHAR* pModelName;
  TCHAR* pBitFilename = NULL;
  ADMXRC3_CARD_INFO cardInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_STATUS status;
  bool_t bShouldConfigure = TRUE;
  size_t n;
  
  assert(NULL != pDesign);
  assert(NULL != pDir);

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Cannot get card information: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    return NULL;
  }

  /* Special case for "Generic ADB3 Device" (Bridge-in-Target FPGA design), which does not expose
  ** target FPGA information because there is VPD that can be retrieved via the PCIe endpoint. */
  if (cardInfo.Model == ADMXRC3_MODEL_GENERIC) {
    const TCHAR* pMessage = _T("Unable to generate bitstream filename for Generic ADB3 Device (ADMXRC3_MODEL_GENERIC)");

    n = _tcslen(pMessage) + 1;
    pBitFilename = malloc(n * sizeof(TCHAR));
    if (NULL == pBitFilename) {
      _tprintf(_T("*** Failed to allocate string buffer for .bit file name.\n"));
      return NULL;
    }
    _tcscpy_s(pBitFilename, n, pMessage);
    if (NULL != pbShouldConfigure) {
      *pbShouldConfigure = FALSE;
    }
    return pBitFilename;
  }

  status = ADMXRC3_GetFpgaInfo(hCard, targetIndex, &fpgaInfo);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Cannot get information for target FPGA %lu(0x%lX): %s\n"),
      (unsigned long)targetIndex, (unsigned long)targetIndex,
      ADMXRC3_GetStatusString(status, TRUE));
    return NULL;
  }
  _tcslwr_s(fpgaInfo.Device, ARRAY_LENGTH(fpgaInfo.Device));

  switch(cardInfo.Model) {
  case ADMXRC3_MODEL_GENERIC:
    _tprintf(_T("*** Cannot construct a .bit filename for a generic reconfigurable computing device.\n"));
    return NULL;

  case ADMXRC3_MODEL_ADMXRC2:
    pModelName = _T("admxrc2");
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    pModelName = _T("adpexrc5t");
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
    pModelName = _T("admxrc6tl");
    break;

  case ADMXRC3_MODEL_ADMXRC6T1:
    pModelName = _T("admxrc6t1");
    break;

  case ADMXRC3_MODEL_ADMXRC6TGE:
    pModelName = _T("admxrc6tge");
    break;

  case ADMXRC3_MODEL_ADMXRC6TGEL:
    pModelName = _T("admxrc6tgel");
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = _T("admxrc6tadv8");
    break;

  case ADMXRC3_MODEL_ADPEXRC6T:
    pModelName = _T("adpexrc6t");
    break;

  case ADMXRC3_MODEL_ADPEXRC6TL:
    pModelName = _T("adpexrc6tl");
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    pModelName = _T("adpexrc6tadv");
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_T:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = _T("adpexrc6tadv");
    break;

  case ADMXRC3_MODEL_ADMXRC7K1:
    pModelName = _T("admxrc7k1");
    break;

  case ADMXRC3_MODEL_ADMXRC6TDA1:
    pModelName = _T("admxrc6tda1");
    break;

  case ADMXRC3_MODEL_ADMXRC7V1:
    pModelName = _T("admxrc7v1");
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    pModelName = _T("admvpx37v2");
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = _T("admpcie7v3");
    break;

  case ADMXRC3_MODEL_ADMPCIEKU3:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = _T("admpcieku3");
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
    pModelName = _T("admxrcku1");
    break;

  case ADMXRC3_MODEL_ADMPCIE8V3:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = _T("admpcie8v3");
    break;

  case ADMXRC3_MODEL_ADMPCIE8K5:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = _T("admpcie8k5");
    break;

  case ADMXRC3_MODEL_ADMXRCKU1AT8:
    pModelName = _T("admxrcku1at8");
    break;

  default:
    _tprintf(_T("*** Cannot construct a .bit file name for unknown model %lu(0x%lX)\n"),
      (unsigned long)cardInfo.Model, (unsigned long)cardInfo.Model);
    return NULL;
  }

  if (NULL != pSdkBitDir) {
    n = _tcslen(pSdkBitDir) + 1 +
        _tcslen(pDir) + 1 +
        _tcslen(pDesign) + 1 +
        _tcslen(pModelName) + 1 +
        _tcslen(fpgaInfo.Device) +
        _tcslen(ext) + 1;
    pBitFilename = malloc(n * sizeof(TCHAR));
    if (NULL == pBitFilename) {
      _tprintf(_T("*** Failed to allocate string buffer for .bit file name.\n"));
      return NULL;
    }
    _tcscpy_s(pBitFilename, n, pSdkBitDir);
    _tcscat_s(pBitFilename, n, nativeSlashStr);
    _tcscat_s(pBitFilename, n, pDir);
    _tcscat_s(pBitFilename, n, nativeSlashStr);
    _tcscat_s(pBitFilename, n, pDesign);
    _tcscat_s(pBitFilename, n, _T("-"));
    _tcscat_s(pBitFilename, n, pModelName);
    _tcscat_s(pBitFilename, n, _T("-"));
    _tcscat_s(pBitFilename, n, fpgaInfo.Device);
    _tcscat_s(pBitFilename, n, ext);
  } else {
    if (bShouldConfigure) {
      _tprintf(_T("+++ WARNING: Could not construct .bit file name because the path to\n"));
      _tprintf(_T("    the SDK bitstream directory could not be determined.\n"));
      _tprintf(_T("    Is the environment variable %s set correctly?\n"), ENVIRONMENT_VARIABLE);
    }
  }
  
  if (NULL != pbShouldConfigure) {
    *pbShouldConfigure = bShouldConfigure;
  }

  return pBitFilename;
}

const TCHAR*
sampleGetFpgaFilename(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  const TCHAR* pDir,
  const TCHAR* pDesign,
  Arguments* pArgs,
  bool_t* pbShouldConfigure)
{
  TCHAR* pBitFilename;
  
  if (NULL != pArgs->bitFile) {
    /* Overrode bit file path construction with '-bitfile' option. */
    if (NULL != pbShouldConfigure) {
      *pbShouldConfigure = TRUE;
    }
    return pArgs->bitFile;
  }

  pBitFilename = sampleConstructBitFilename(hCard, targetIndex, pDir, pDesign, pArgs->bitPath, pbShouldConfigure);
  if (NULL == pBitFilename) {
    if (*pbShouldConfigure) {
      _tprintf(
        _T("*** The path to the FPGA bitstream directory cannot be determined via the\n")
        _T("    %s environment variable and has not been specified via the /sdk or\n")
        _T("    /bitdir switches. Please correct this and try again.\n"),
        ENVIRONMENT_VARIABLE);
    }
    return NULL;
  }

  pArgs->bitFile = pBitFilename;

  return pBitFilename;
}
